@@ -4,7 +4,6 @@ import datetime as dt
 from decimal import Decimal
 
 from trytond.config import config
-from trytond.transaction import Transaction
 from trytond.pool import Pool, PoolMeta
 
 from trytond.modules.product import round_price
@@ -93,7 +92,6 @@ class Product(metaclass=PoolMeta):
     def recompute_cost_price_fifo(self, start=None):
         pool = Pool()
         Move = pool.get('stock.move')
-        Currency = pool.get('currency.currency')
         Uom = pool.get('product.uom')
         Revision = pool.get('product.cost_price.revision')
 
@@ -143,27 +141,21 @@ class Product(metaclass=PoolMeta):
             return move.from_location.type == 'storage'
 
         def compute_fifo_cost_price(quantity, date):
-            fifo_moves = self.get_fifo_move(
-                float(quantity),
-                date=current_moves[-1].effective_date)
+            fifo_moves = self.get_fifo_move(float(quantity), date=date)
 
             cost_price = Decimal(0)
             consumed_qty = 0
             for move, move_qty in fifo_moves:
                 consumed_qty += move_qty
-                if move.from_location.type in {'supplier', 'production'}:
-                    with Transaction().set_context(date=move.effective_date):
-                        unit_price = Currency.compute(
-                            move.currency, move.unit_price,
-                            move.company.currency, round=False)
-                    unit_price = Uom.compute_price(
-                        move.uom, unit_price, move.product.default_uom)
-                else:
-                    unit_price = move.cost_price or 0
-                cost_price += unit_price * Decimal(str(move_qty))
+                cost_price += move.get_cost_price() * Decimal(str(move_qty))
             if consumed_qty:
                 return round_price(cost_price / Decimal(str(consumed_qty)))
 
+        # For each day, process the incoming moves first
+        # in order to keep quantity positive where possible
+        # We do not re-browse because we expect only small changes
+        moves = sorted(moves, key=lambda m: (
+                m.effective_date, out_move(m), m.id))
         current_moves = []
         current_out_qty = 0
         current_cost_price = cost_price
@@ -186,14 +178,14 @@ class Product(metaclass=PoolMeta):
                             m for m in out_moves
                             if m.cost_price != fifo_cost_price],
                         dict(cost_price=fifo_cost_price))
-                    if quantity:
+                    if quantity > 0 and quantity + current_out_qty >= 0:
                         cost_price = (
                             ((current_cost_price * (
                                         quantity + current_out_qty))
                                 - (fifo_cost_price * current_out_qty))
                             / quantity)
                     else:
-                        cost_price = Decimal(0)
+                        cost_price = current_cost_price
                     current_cost_price = round_price(cost_price)
                 current_moves.clear()
                 current_out_qty = 0
@@ -206,15 +198,7 @@ class Product(metaclass=PoolMeta):
             if move.from_location.type == 'storage':
                 qty *= -1
             if in_move(move):
-                if move.from_location.type in {'supplier', 'production'}:
-                    with Transaction().set_context(date=move.effective_date):
-                        unit_price = Currency.compute(
-                            move.currency, move.unit_price,
-                            move.company.currency, round=False)
-                    unit_price = Uom.compute_price(
-                        move.uom, unit_price, self.default_uom)
-                else:
-                    unit_price = cost_price
+                unit_price = move.get_cost_price(product_cost_price=cost_price)
                 if quantity + qty > 0 and quantity >= 0:
                     cost_price = (
                         (cost_price * quantity) + (unit_price * qty)
@@ -246,8 +230,6 @@ class Product(metaclass=PoolMeta):
                     ((cost_price * (quantity + current_out_qty))
                         - (fifo_cost_price * current_out_qty))
                     / quantity)
-            else:
-                cost_price = Decimal(0)
         for revision in revisions:
             cost_price = revision.get_cost_price(cost_price)
         return cost_price
